/*
 * SkyTimes.java
 *
 * Created on 12. Mrz 2008, 00:41
 */

/**
 *
 * @author paddy
 */

package PGRWetter;

import java.util.Calendar;
import java.util.TimeZone;


public class SkyTimes {
    
    private int DeltaT = 65;    
    private double DEG = Math.PI / 180.0;
    private double RAD = 180.0 / Math.PI;

    /** Creates a new instance of SkyTimes */
    public SkyTimes() {
    }
    
// return integer value, closer to 0
    private int Int(double x) {
	if (x < 0) {
	    return( (int) Math.ceil(x) );
	} else {
	    return( (int) Math.floor(x) );
	}
    }
    
    private double frac(double x) {
	return( x - (Math.floor(x)) );
    }
    
    private double Mod(double a, double b) {
	return( a - Math.floor(a / b) * b);
    }
    
// Modulo PI
    private double Mod2Pi(double x) {
	return( Mod(x * 1.0, 2.0 * Math.PI) );
    }
    
    private String HHMM(double hh) {
	String timeString = "";
	int h = Int(hh);
	double m = frac(hh) * 60;
	if (m >= 59.5) {
	    h++;
	    m -= 60;
	}
	m = Math.round(m);
	if (h < 10){ timeString = "0" + h; } else { timeString = "" + h; }
	timeString = timeString + ":";
	if (m < 10){ timeString = timeString + "0" + (int)m; } else { timeString = timeString + (int)m;	}
	return(timeString);
    }
    
// Calculate Julian date: valid only from 1.3.1901 to 28.2.2100
    private double CalcJD(int day, int month, int year) {
	double jd = 2415020.5 - 64; // 1.1.1900 - correction of algorithm
	if (month <= 2) {
	    year--;
	    month += 12;
	}
	jd += Int( (year - 1900) * 365.25 );
	jd += Int( 30.6001 * (1 + month) );
	return(jd + day);
    }
    
// Julian Date to Greenwich Mean Sidereal Time
    private double GMST(double JD) {
	double UT = frac(JD - 0.5) * 24; // UT in hours
	JD = Math.floor(JD - 0.5) + 0.5;   // JD at 0 hours UT
	double T = (JD - 2451545.0) / 36525.0;
	double T0 = 6.697374558 + T * (2400.051336 + T * 0.000025862);
	return( Mod(T0 + UT * 1.002737909, 24) );
    }
    
    
// Convert Greenweek mean sidereal time to UT
    private double GMST2UT(double JD, double gmst) {
	JD = Math.floor(JD - 0.5) + 0.5;   // JD at 0 hours UT
	double T = (JD - 2451545.0) / 36525.0;
	double T0 = Mod(6.697374558 + T * (2400.051336 + T * 0.000025862), 24);
	//double UT = 0.9972695663*Mod((gmst-T0), 24.0);
	double UT = 0.9972695663 * (gmst-T0);
	return(UT);
    }
    
// Local Mean Sidereal Time, geographical longitude in radians, East is positive
    private double GMST2LMST(double gmst, double lon) {
	double lmst = Mod(gmst + RAD * lon / 15, 24.0);
	return( lmst );
    }
    
// Transform ecliptical coordinates (lon/lat) to equatorial coordinates (RA/dec)
    private sunCoorObject Ecl2Equ(sunCoorObject coor, double TDT) {
	double T = (TDT - 2451545.0) / 36525.; // Epoch 2000 January 1.5
	double eps = (23.0 + (26 + 21.45 / 60) /60 + T * (-46.815 + T * (-0.0006 + T * 0.00181) ) / 3600 ) * DEG;
	double coseps = Math.cos(eps);
	double sineps = Math.sin(eps);
	
	double sinlon = Math.sin(coor.lon);
	coor.ra  = Mod2Pi( Math.atan2( (sinlon*coseps-Math.tan(coor.lat)*sineps), Math.cos(coor.lon) ) );
	coor.dec = Math.asin( Math.sin(coor.lat)*coseps + Math.cos(coor.lat)*sineps*sinlon );
	
	return coor;
    }
    
// Transform equatorial coordinates (RA/Dec) to horizonal coordinates (azimuth/altitude)
// Refraction is ignored
    private sunCoorObject Equ2Altaz(sunCoorObject coor, double TDT, double geolat, double lmst) {
	double cosdec = Math.cos(coor.dec);
	double sindec = Math.sin(coor.dec);
	double lha = lmst - coor.ra;
	double coslha = Math.cos(lha);
	double sinlha = Math.sin(lha);
	double coslat = Math.cos(geolat);
	double sinlat = Math.sin(geolat);
	
	double N = -cosdec * sinlha;
	double D = sindec * coslat - cosdec * coslha * sinlat;
	coor.az = Mod2Pi( Math.atan2(N, D) );
	coor.alt = Math.asin( sindec * sinlat + cosdec * coslha * coslat );
	
	return coor;
    }
    
// Calculate cartesian from polar coordinates
    private cartesCoorObject EquPolar2Cart(double lon, double lat, double distance ) {
	cartesCoorObject cart = new cartesCoorObject();
	double rcd = Math.cos(lat) * distance;
	cart.x = rcd*Math.cos(lon);
	cart.y = rcd*Math.sin(lon);
	cart.z = distance * Math.sin(lat);
	return(cart);
    }
    
// Calculate observers cartesian equatorial coordinates (x,y,z in celestial frame)
// from geodetic coordinates (longitude, latitude, height above WGS84 ellipsoid)
// Currently only used to calculate distance of a body from the observer
    private cartesCoorObject Observer2EquCart(double lon, double lat, double height, double gmst ) {
	double flat = 298.257223563;        // WGS84 flatening of earth
	double aearth = 6378.137;           // GRS80/WGS84 semi major axis of earth ellipsoid
	cartesCoorObject cart = new cartesCoorObject();
	// Calculate geocentric latitude from geodetic latitude
	double co = Math.cos(lat);
	double si = Math.sin(lat);
	double fl = 1.0 - 1.0 / flat;
	fl = fl * fl;
	si = si * si;
	
	double u = 1.0 / Math.sqrt(co * co + fl * si);
	double a = aearth * u + height;
	double b = aearth * fl * u + height;
	double radius = Math.sqrt(a * a * co * co + b * b * si); // geocentric distance from earth center
	
	cart.y = Math.acos(a * co / radius); // geocentric latitude, rad
	cart.x = lon; // longitude stays the same
	if (lat < 0.0) { cart.y = -cart.y; } // adjust sign
	cart = EquPolar2Cart( cart.x, cart.y, radius ); // convert from geocentric polar to geocentric cartesian, with regard to Greenwich
	// rotate around earth's polar axis to align coordinate system from Greenwich to vernal equinox
	double x = cart.x;
	double y = cart.y;
	double rotangle = gmst / 24 * 2 * Math.PI; // sideral time gmst given in hours. Convert to radians
	cart.x = x * Math.cos(rotangle) - y * Math.sin(rotangle);
	cart.y = x * Math.sin(rotangle) + y * Math.cos(rotangle);
	cart.radius = radius;
	cart.lon = lon;
	cart.lat = lat;
	return(cart);
    }
    
// Calculate coordinates for Sun
// Coordinates are accurate to about 10s (right ascension)
// and a few minutes of arc (declination)
    private sunCoorObject SunPosition(double TDT, double geolat, double lmst) {
	double D = TDT - 2447891.5;
	
	double eg = 279.403303 * DEG;
	double wg = 282.768422 * DEG;
	double e  = 0.016713;
	double a  = 149598500; // km
	double diameter0 = 0.533128 * DEG; // angular diameter of Moon at a distance
	
	double MSun = 360 * DEG /365.242191 * D + eg - wg;
	double nu = MSun + 360.0 * DEG / Math.PI * e * Math.sin(MSun);
	
	sunCoorObject sunCoor = new sunCoorObject();
	sunCoor.lon =  Mod2Pi(nu + wg);
	sunCoor.lat = 0;
	sunCoor.anomalyMean = MSun;
	
	sunCoor.distance = ( 1 - Math.sqrt(e) ) / ( 1 + e * Math.cos(nu) ); // distance in astronomical units
	sunCoor.diameter = diameter0/sunCoor.distance; // angular diameter in radians
	sunCoor.distance *= a;                         // distance in km
	sunCoor.parallax = 6378.137 / sunCoor.distance;  // horizonal parallax
	
	sunCoor = Ecl2Equ(sunCoor, TDT);
	
	// Calculate horizonal coordinates of sun, if geographic positions is given
	if (geolat > 0 && lmst > 0) {
	    sunCoor = Equ2Altaz(sunCoor, TDT, geolat, lmst);
	}
	
	//sunCoor.sign = Sign(sunCoor.lon);
	return sunCoor;
    }
    
// Rough refraction formula using standard atmosphere: 1015 mbar and 10C
// Input true altitude in radians, Output: increase in altitude in degrees
    private double Refraction(double alt) {
	double altdeg = alt*RAD;
	if (altdeg < -2 || altdeg >= 90){
	    return(0);
	}
	
	double pressure = 1015;
	double temperature = 10;
	if (altdeg > 15){
	    return( 0.00452 * pressure / ( (273 + temperature) * Math.tan(alt) ) );
	}
	
	double y = alt;
	double D = 0.0;
	double P = (pressure - 80.0) / 930.0;
	double Q = 0.0048 * (temperature - 10.0);
	double y0 = y;
	double D0 = D;
	
	for (int i = 0; i < 3; i++) {
	    double N = y + (7.31 / (y + 4.4));
	    N = 1.0 / Math.tan(N * DEG);
	    D = N * P / (60.0 + Q * (N + 39.0));
	    N = y - y0;
	    y0 = D - D0 - N;
	    if ((N != 0.0) && (y0 != 0.0)) {
		N = y - N * (alt + D - y) / y0;
	    } else {
		N = alt + D;
	    }
	    y0 = y;
	    D0 = D;
	    y = N;
	}
	return( D ); // Hebung durch Refraktion in radians
    }
    
    
// returns Greenwich sidereal time (hours) of time of rise
// and set of object with coordinates coor.ra/coor.dec
// at geographic position lon/lat (all values in radians)
// Correction for refraction and semi-diameter/parallax of body is taken care of in function RiseSet
// h is used to calculate the twilights. It gives the required elevation of the disk center of the sun
    private riseTransSetObject GMSTRiseSet(sunCoorObject coor, double lon, double lat, double h) {
	h = (h == 0) ? 0. : h; // set default value
	riseTransSetObject riseset = new riseTransSetObject();
	//  double tagbogen = Math.acos(-Math.tan(lat)*Math.tan(coor.dec)); // simple formula if twilight is not required
	double tagbogen = Math.acos( (Math.sin(h) -
		Math.sin(lat) * Math.sin(coor.dec) ) /
		(Math.cos(lat) * Math.cos(coor.dec)));
	
	riseset.transit =        RAD / 15 * (          +coor.ra -lon);
	riseset.rise    = 24.0 + RAD / 15 * (-tagbogen +coor.ra -lon); // calculate GMST of rise of object
	riseset.set     =        RAD / 15 * (+tagbogen +coor.ra -lon); // calculate GMST of set of object
	
	// using the modulo function Mod, the day number goes missing. This may get a problem for the moon
	riseset.transit = Mod(riseset.transit, 24);
	riseset.rise    = Mod(riseset.rise, 24);
	riseset.set     = Mod(riseset.set, 24);
	
	return(riseset);
    }
    
    
// Find GMST of rise/set of object from the two calculates
// (start)points (day 1 and 2) and at midnight UT(0)
    private double InterpolateGMST(double gmst0, double gmst1, double gmst2, double timefactor) {
	return( (timefactor*24.07*gmst1- gmst0*(gmst2-gmst1)) / (timefactor*24.07+gmst1-gmst2) );
    }
    
    
// JD is the Julian Date of 0h UTC time (midnight)
    private riseTransSetObject RiseSet(double jd0UT, sunCoorObject coor1, sunCoorObject coor2, double lon, double lat, double timeinterval, double altitude) {
	// altitude of sun center: semi-diameter, horizontal parallax and (standard) refraction of 34'
	double alt = 0.0; // calculate
	altitude = (altitude == 0) ? 0.0 : altitude; // set default value
	
	// true height of sun center for sunrise and set calculation. Is kept 0 for twilight (ie. altitude given):
	if (altitude == 0){
	    alt = 0.5 * coor1.diameter - coor1.parallax + 34.0 / 60 * DEG;
	}
	
	riseTransSetObject rise1 = GMSTRiseSet(coor1, lon, lat, altitude);
	riseTransSetObject rise2 = GMSTRiseSet(coor2, lon, lat, altitude);
	
	riseTransSetObject rise = new riseTransSetObject();
	
	// unwrap GMST in case we move across 24h -> 0h
	if ((rise1.transit > rise2.transit) && (Math.abs(rise1.transit - rise2.transit) > 18)){ rise2.transit += 24; }
	if ((rise1.rise    > rise2.rise   ) && (Math.abs(rise1.rise    - rise2.rise)    > 18)){ rise2.rise    += 24; }
	if ((rise1.set     > rise2.set    ) && (Math.abs(rise1.set     - rise2.set)     > 18)){ rise2.set     += 24; }
	
	double T0 = GMST(jd0UT);
	//  var T02 = T0-zone*1.002738; // Greenwich sidereal time at 0h time zone (zone: hours)
	
	// Greenwich sidereal time for 0h at selected longitude
	double T02 = T0 - lon * RAD / 15 * 1.002738;
	if (T02 < 0){ T02 += 24; }
	
	if (rise1.transit < T02){ rise1.transit += 24; rise2.transit += 24; }
	if (rise1.rise    < T02){ rise1.rise    += 24; rise2.rise    += 24; }
	if (rise1.set     < T02){ rise1.set     += 24; rise2.set     += 24; }
	
	// Refraction and Parallax correction
	double decMean = 0.5 * (coor1.dec + coor2.dec);
	double psi = Math.acos(Math.sin(lat) / Math.cos(decMean));
	double y = Math.asin(Math.sin(alt) / Math.sin(psi));
	double dt = 240 * RAD * y / Math.cos(decMean) / 3600; // time correction due to refraction, parallax
	
	rise.transit = GMST2UT( jd0UT, InterpolateGMST( T0, rise1.transit, rise2.transit, timeinterval)      );
	rise.rise    = GMST2UT( jd0UT, InterpolateGMST( T0, rise1.rise,    rise2.rise,    timeinterval) - dt );
	rise.set     = GMST2UT( jd0UT, InterpolateGMST( T0, rise1.set,     rise2.set,     timeinterval) + dt );
	
	return(rise);
    }
    
    
// Find (local) time of sunrise and sunset, and twilights
// JD is the Julian Date of 0h local time (midnight)
// Accurate to about 1-2 minutes
// recursive: 1 - calculate rise/set in UTC in a second run
// recursive: 0 - find rise/set on the current local day. This is set when doing the first call to this function
    private riseTransSetObject SunRise(double JD, double deltaT, double lon, double lat) {
	double jd0UT = Math.floor(JD - 0.5) + 0.5;   // JD at 0 hours UT
	sunCoorObject coor1 = SunPosition(jd0UT + 0 * deltaT / 24 / 3600, 0, 0);
	sunCoorObject coor2 = SunPosition(jd0UT + 1 + 0 * deltaT / 24 /3600, 0, 0); // calculations for next day's UTC midnight
	
	riseTransSetObject risetemp = new riseTransSetObject();
	riseTransSetObject rise = new riseTransSetObject();
	
	// rise/set time in UTC.
	rise = RiseSet(jd0UT, coor1, coor2, lon, lat, 1, coor1.alt);
	
	rise.transit = Mod(rise.transit, 24);
	rise.rise    = Mod(rise.rise   , 24);
	rise.set     = Mod(rise.set    , 24);
	
	// Twilight calculation
	// civil twilight time in UTC.
	risetemp = RiseSet(jd0UT, coor1, coor2, lon, lat, 1, -6.0 * DEG);
	rise.civilTwilightMorning = Mod(risetemp.rise, 24);
	rise.civilTwilightEvening = Mod(risetemp.set , 24);
	
	// nautical twilight time in UTC.
	risetemp = RiseSet(jd0UT, coor1, coor2, lon, lat, 1, -12.0 * DEG);
	rise.nauticalTwilightMorning = Mod(risetemp.rise, 24);
	rise.nauticalTwilightEvening = Mod(risetemp.set , 24);
	
	// astronomical twilight time in UTC.
	risetemp = RiseSet(jd0UT, coor1, coor2, lon, lat, 1, -18.0 * DEG);
	rise.astronomicalTwilightMorning = Mod(risetemp.rise, 24);
	rise.astronomicalTwilightEvening = Mod(risetemp.set , 24);
	
	return( rise );
    }
    
    public String[] getRiseSetTimes(long stationDateLong, double latitude, double longitude, int zenithMode) {
	Calendar calcCal = Calendar.getInstance(TimeZone.getTimeZone("Etc/UTC"));
	calcCal.setTimeInMillis(stationDateLong);
	
	double JD0 = CalcJD( Calendar.DAY_OF_MONTH, Calendar.MONTH + 1, Calendar.YEAR );
	double JD  = JD0 + ( Calendar.HOUR_OF_DAY + (Calendar.MINUTE / 60.0) + (Calendar.SECOND / 3600)) / 24.0;
	double TDT = JD + DeltaT / 24.0 / 3600.0;
	double lat = latitude * DEG; // geodetic latitude of observer on WGS84
	double lon = longitude * DEG; // latitude of observer
	
//	System.out.println( "... stationDateLong: " + stationDateLong );
//	System.out.println( "... calcCal.getTime().toString(): " + calcCal.getTime().toString() );
//	System.out.println( "... calcCal.getTimeZone().getRawOffset(): " + calcCal.getTimeZone().getRawOffset() );
//	System.out.println( "... latitude: " + latitude );
//	System.out.println( "... lat: " + lat );
//	System.out.println( "... longitude: " + longitude );
//	System.out.println( "... lon: " + lon );

	riseTransSetObject sunRise = SunRise(JD0, DeltaT, lon, lat);
	
//	System.out.println( "... HHMM(sunRise.rise): " + HHMM(sunRise.rise) );	
//	System.out.println( "... HHMM(sunRise.set): " + HHMM(sunRise.set) );
//	System.out.println( "");
//	System.out.println( "... HHMM(sunRise.civilTwilightMorning): " + HHMM(sunRise.civilTwilightMorning) );
//	System.out.println( "... HHMM(sunRise.civilTwilightEvening): " + HHMM(sunRise.civilTwilightEvening) );
//	System.out.println( "");
//	System.out.println( "... HHMM(sunRise.nauticalTwilightMorning): " + HHMM(sunRise.nauticalTwilightMorning) );
//	System.out.println( "... HHMM(sunRise.nauticalTwilightEvening): " + HHMM(sunRise.nauticalTwilightEvening) );
//	System.out.println( "");
//	System.out.println( "... HHMM(sunRise.astronomicalTwilightMorning): " + HHMM(sunRise.astronomicalTwilightMorning) );
//	System.out.println( "... HHMM(sunRise.astronomicalTwilightEvening): " + HHMM(sunRise.astronomicalTwilightEvening) );
	
	String[] returnTimesStringArray = {"", ""};	
	if(zenithMode == 1){
	    returnTimesStringArray[0] = HHMM(sunRise.rise);
	    returnTimesStringArray[1] = HHMM(sunRise.set);
	} else if(zenithMode == 2){
	    returnTimesStringArray[0] = HHMM(sunRise.civilTwilightMorning);
	    returnTimesStringArray[1] = HHMM(sunRise.civilTwilightEvening);
	} else if(zenithMode == 3){
	    returnTimesStringArray[0] = HHMM(sunRise.nauticalTwilightMorning);
	    returnTimesStringArray[1] = HHMM(sunRise.nauticalTwilightEvening);
	} else if(zenithMode == 4){
	    returnTimesStringArray[0] = HHMM(sunRise.astronomicalTwilightMorning);
	    returnTimesStringArray[1] = HHMM(sunRise.astronomicalTwilightEvening);
	}
	return returnTimesStringArray;
    }
    
}
