/*
 * Sun.java
 *
 * Created on 27. Februar 2008, 06:42
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

/**
 *
 * @author paddy
 */

package PGRWetter;

import java.util.Calendar;
import java.util.TimeZone;

/**
 * Liefert die Uhrzeiten fr Sonnenauf- und -untergang.
 * Benutzt dafr SunTimes (com.web_tomorrow.utils.suntimes)
 * von Kevin Boone/Web-Tomorrow
 */
public class Sun {
    /**
     * Latitude zur Berechnung
     */
    private static String latitude;
    /**
     * Longitude zur Berechnung
     */
    private static String longitude;
    /**
     * Datum des zu berechnenden Punktes
     */
    private static long stationDateLong;
    /**
     * welcher Zenith:
     * 1 = ZENITH (true rise/set)
     * 2 = CIVIL_ZENITH (brgerliche Dmmerung bzw. zivile Dmmerung)
     * 3 = NAUTICAL_ZENITH (nautische Dmmerung oder mittlere Dmmerung; schliet an die brgerliche Dmmerung an)
     * 4 = ASTRONOMICAL_ZENITH (astronomische Dmmerung; folgt der nautischen Dmmerung)
     */
    private static int zenithMode;
    /**
     * doubleWert des Zeniths
     */
    private static double zenithDouble;
    /**
     * Calender mit dem Datum des zu berechnende Punktes
     */
    private static Calendar stationCal = Calendar.getInstance();
    /**
     * Calendar mit dem Datum des jeweiligen Rckgabewertes
     */
    private static Calendar returnCal = Calendar.getInstance();
    
    private String[] sunTimesStringArray = {"", ""};
    
    /**
     * Konstruktor der Klasse
     * @param latitude Latitude des zu berechnenden Punktes
     * @param longitude Longitude des zu berechnenden Punktes
     * @param stationDateLong Datum des zu berechenden Punktes
     */
    public Sun(String latitude, String longitude, long stationDateLong, int zenithMode){
	Sun.latitude = latitude;
	Sun.longitude = longitude;
	Sun.stationDateLong = stationDateLong;
	Sun.zenithMode = zenithMode;
	
	stationCal.setTimeZone( TimeZone.getTimeZone("Etc/UTC") );
	returnCal.setTimeZone( TimeZone.getTimeZone("Etc/UTC") );
	stationCal.setTimeInMillis(stationDateLong);

	SkyTimes skyTimes = new SkyTimes();
	sunTimesStringArray = skyTimes.getRiseSetTimes(stationDateLong,
		latitudeToDouble(latitude),
		longitudeToDouble(longitude),
		zenithMode);
    }
    
    /**
     * Gibt das Datum mit der Uhrzeit des Sonnenaufgangs zurck
     * @return der Timestamp des Sonnenaufgangs (long)
     */
    public final long computeSunRise() {
	String[] sunRiseString = sunTimesStringArray[0].toString().split(":");
	returnCal.set(stationCal.get(Calendar.YEAR),
		stationCal.get(Calendar.MONTH),
		stationCal.get(Calendar.DAY_OF_MONTH),
		Integer.parseInt(sunRiseString[0]),
		Integer.parseInt(sunRiseString[1]),
		0);
	return returnCal.getTimeInMillis();
    }
    
    /**
     * Gibt ein Datum mit der Uhrzeit des Sonnenuntergangs zurck
     * @return der Timestamp mit der Uhrzeit des Sonnenuntergangs
     */
    public final long computeSunSet() {
	String[] sunSetString = sunTimesStringArray[1].toString().split(":");
	returnCal.set(stationCal.get(Calendar.YEAR),
		stationCal.get(Calendar.MONTH),
		stationCal.get(Calendar.DAY_OF_MONTH),
		Integer.parseInt(sunSetString[0]),
		Integer.parseInt(sunSetString[1]),
		0);
	return returnCal.getTimeInMillis();
    }
    
/*
 * Converts latitude in format DD-MMH, where DD is degrees, MM is minutes,
 * and H is Hemisphere (N for North, or S for South) to a floating point number.
 *
 * For example: 27-00S to -27.0
 *
 * Does not currently handle seconds.
 */
    /**
     * Macht aus einem Latitude als String den Long-Wert
     * z.B. aus "27-00S" -27.0
     * @param latString die Latitude als String im Format "##-##N" oder "##-##S"
     * @return die Latitude als Double
     */
    private static double latitudeToDouble(String latString){
	double result = 0;
	double dd = Double.parseDouble(latString.substring(0, 2));
	double mm = Double.parseDouble(latString.substring(3, 5));
	result = dd + (mm / 100);
	if(latString.contains("S")){
	    result *= -1;
	}
	return result;
    }
    /**
     * Macht aus einem Longitude als String den Long-Wert
     * z.B. aus "010-00E" 1.0
     * @param longString die Longitude als String im Format "###-##W" oder "###-##E"
     * @return die Longitude als Double
     */
    private static double longitudeToDouble(String longString){
	double result = 0;
	double ddd = Double.parseDouble(longString.substring(0, 3));
	double mm = Double.parseDouble(longString.substring(4, 6));
	result = ddd + (mm / 100);
	if(longString.contains("W")){
	    result *= -1;
	}
	return result;
    }
}
