/*
 * stationsWorker.java
 *
 * Created on 24. Februar 2008, 23:45
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

/**
 *
 * @author paddy
 */

package PGRWetter;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;

public class stationsWorker {

    private static TreeMap<String, String> stationsFull = new TreeMap<String, String>();
    private static String actStationID = "";

    private final static DecimalFormat decFormat = new DecimalFormat("#,##0.0");
    private final static DecimalFormat decVKFormat = new DecimalFormat("#,##0");
    public final static SimpleDateFormat sdfUTC = new SimpleDateFormat("HH:mm 'Uhr' z");
    public final static SimpleDateFormat sdfLocal = new SimpleDateFormat("HH:mm 'Uhr' z");

    private static PGRWetter pgrWetter;
    
    public static int zenithModeInt = 2;
    public static String actStationLat = "";
    public static String actStationLong = "";
    public static String stationStringToParse = "";
    public static String decodedStringToParse = "";

    private final static String[][] wind_directions = {
	{ "0",   "N"   },
	{ "12",  "NNE" },
	{ "33",  "NE"  },
	{ "57",  "ENE" },
	{ "79",  "E"   },
	{ "102", "ESE" },
	{ "124", "SE"  },
	{ "147", "SSE" },
	{ "169", "S"   },
	{ "192", "SSW" },
	{ "214", "SW"  },
	{ "237", "WSW" },
	{ "259", "W"   },
	{ "282", "WNW" },
	{ "304", "NW"  },
	{ "327", "NNW" },
	{ "349", "N"   },
	{ "360", "N"   }
    };

    private static int info_cloudsInt = 0;
    private static float info_windMphFloat = 0;
    private static float info_tempCFloat = 0f;
    private static float info_dewCFloat  = 0f;
    private static boolean info_heavyBool = false;
    private static boolean info_stationNeedsMaintBool = false;

    private Pattern CoverRegExp	    = Pattern.compile("^(FEW|SCT|BKN|OVC|SKC|CLR|CAVOK)([0-9]{3})?(?:TCU|CB)?$");
    private Pattern CurrentRegExp   = Pattern.compile("^(\\+|-|VC)?([A-Z]{2,4})$");
    private Pattern WindRegExp	    = Pattern.compile("^([0-9]{3}|VRB)([0-9]{2,3})(?:G([0-9]{2,3}))?(KT|KMH|MPS)$");
    private Pattern VisRegExp	    = Pattern.compile("^([0-9]{4})$");
    //private Pattern VisRegExp	    = Pattern.compile("^([0-9]{1,2})SM$");
    //private Pattern VisFracRegExp   = Pattern.compile("^1/(2|4)SM$");
    private Pattern TempRegExp	    = Pattern.compile("^(M)?([0-9]{2})/(?:(M)?([0-9]{2}))?$");
    private Pattern TimeRegExp	    = Pattern.compile("^([0-9]{2}:[0-9]{2})$");
    private Pattern DateRegExp	    = Pattern.compile("^([0-9]{4}/[0-9]{2}/[0-9]{2})$");
    private Pattern PressRegExp	    = Pattern.compile("^([AQ])([0-9]{4})$");
    private Pattern TempTenRegExp   = Pattern.compile("^T([01][0-9]{3})([01][0-9]{3})$");
    private Pattern StatNeedsMaint  = Pattern.compile("$");

    private Pattern findLat  = Pattern.compile("^([0-9]{2})(\\-)?([0-9]{2})(N|S)$");
    private Pattern findLong  = Pattern.compile("^([0-9]{3})(\\-)?([0-9]{2})(E|W)$");
    //private Pattern findHeight  = Pattern.compile("^([0-9]+)(M)$");

    /**
     * Creates a new instance of stationsWorker
     * @param pgrWetter
     */
    @SuppressWarnings("static-access")
    public stationsWorker(PGRWetter pgrWetter){
	//System.out.println("+++ pgrWetter: " + pgrWetter + " +++");
	stationsWorker.pgrWetter = pgrWetter;
	sdfUTC.setTimeZone(TimeZone.getTimeZone("Etc/UTC"));
	sdfLocal.setTimeZone(TimeZone.getDefault());
    }

    public synchronized TreeMap<String, String> getStationsMap() {
	createStationsFull();
	TreeMap<String, String> stationsMap = new TreeMap<String, String>();
	try {
            @SuppressWarnings("static-access")
	    URL urlobj = new URL(pgrWetter.metarStationsString);
	    HttpURLConnection uc = (HttpURLConnection)urlobj.openConnection();
	    uc.setUseCaches(false);

	    DataInputStream is = new DataInputStream( uc.getInputStream() );

	    HTMLEditorKit hKit = new HTMLEditorKit();
	    HTMLDocument hDoc = new HTMLDocument();

	    hKit.read(is, hDoc, 0);

	    HTMLDocument.Iterator it = hDoc.getIterator(HTML.Tag.A);

	    while(it.isValid()){
		AttributeSet attSet = it.getAttributes();
		String hrefAttr = attSet.getAttribute(HTML.Attribute.HREF).toString();
		if(hrefAttr.substring(hrefAttr.length() - 4).equalsIgnoreCase(".txt")){
		    String stationID = hrefAttr.substring(0, hrefAttr.length() - 4);
		    if( stationsFull.containsKey( stationID ) ){
			stationsMap.put(stationID, stationsFull.get(stationID));
		    } else {
			stationsMap.put(stationID, "unbekannt");
		    }
		}
		it.next();
	    }

	    is.close();
	    pgrWetter.offline = false;
	} catch (MalformedURLException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getStationsMap() MalformedURLException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	} catch (IOException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getStationsMap() IOException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	} catch (BadLocationException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getStationsMap() BadLocationException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	}
	return stationsMap;
    }

    @SuppressWarnings("static-access")
    public synchronized void updateAllWeatherInfos(String stationID) {
	pgrWetter.actInfo_CurrentVect.clear();
	pgrWetter.actInfo_CoverVect.clear();
	pgrWetter.actInfo_TheWeatherString = "";

	parseMetar(stationID);
	pgrWetter.actInfo_DecodedString = getDecodedString(stationID);

	//System.out.println("+++ updateAllWeatherInfos()");
	//System.out.println("+++ stationStringToParse: " + stationStringToParse + " +++");
    }

    @SuppressWarnings("static-access")
    public synchronized void parseMetar(String stationID){
	// die Daten der Station holen
	actStationID = stationID;
	stationStringToParse = getStringToParseString(stationID);

	info_cloudsInt = 0;
	info_windMphFloat = 0;
	info_tempCFloat = 0f;
	info_dewCFloat  = 0f;
	info_heavyBool = false;
	info_stationNeedsMaintBool = false;
	pgrWetter.actInfo_TheWeatherString = "";

	// die Datenfelder in ein Stringarray packen und parsen
	//String[] parseData = stationStringToParse.trim().split("\\s+"); // ### warum schneidet das das Minuszeichen ab?
	String[] parseData = stationStringToParse.trim().split("[ ]+");

	//System.out.println("### stationStringToParse: '" + stationStringToParse + "' ###");
	boolean found = false;
	boolean beforeRemark = true;
	for(int i = 0; i < parseData.length; i++){
	    String it = parseData[i];
	    //System.out.println("### parseData[" + i + "]: '" + parseData[i] + "' ###");

	    // Don't try to parse the ICAO location code
	    if ( (!found) && ( it.equals(stationID)) ) {
		found = true;
		continue;
	    }
	    if (it.equals("RMK")) {
		beforeRemark = false;
		continue;
	    }
	    if (beforeRemark) {
		if (parseDate(it)){ continue; }
		if (parseTime(it)){ continue; }
		if (parseWindSpeed(it)){ continue; }
		if (parseVisibility(it)){ continue; } // Note, pass in iterator.
		if (parseTemperature(it)){ continue; }
		if (parsePressure(it)){ continue; }
		if (parseCover(it)){ continue; }
		if (parseCurrent(it)){ continue; }
	    } else {
		if (parseTemperatureTenths(it)){ continue; }
		if (parseStationNeedsMaintenance(it)){ continue; }
	    }

	}

	if (!info_stationNeedsMaintBool){
	    calcTemperatureVariables();
	    calcWindChill();
	    calcCurrentIcon();
	}
    }

    private float defineE(float t){
	return (float) Math.pow(10, 7.5 * t / (237.7 + t));
    }
    private float defineSQR(float a){
	return ( a * a );
    }

    @SuppressWarnings("static-access")
    private void calcTemperatureVariables(){
	//System.out.println("+++ calcTemperatureVariables() +++");
	float fRelHumidity = defineE(info_dewCFloat) / defineE(info_tempCFloat) * 100;
	pgrWetter.actInfo_RelHumidityString = removeTrailingDotZero(decFormat.format(fRelHumidity)) + "%";

	float fHeatIndex = 0f;
	float tempF = TEMP_C_TO_F(info_tempCFloat);
	if (tempF >= 80) {
	    fHeatIndex = -42.379f +
		    ( 2.04901523f * tempF ) +
		    ( 10.14333127f * fRelHumidity ) -
		    ( 0.22475541f * tempF * fRelHumidity ) -
		    ( 0.00683783f * defineSQR(tempF) ) -
		    ( 0.05481717f * defineSQR(fRelHumidity) ) +
		    ( 0.00122874f * defineSQR(tempF) * fRelHumidity ) +
		    ( 0.00085282f * tempF * defineSQR(fRelHumidity) ) -
		    ( 0.00000199f * defineSQR(tempF) * defineSQR(fRelHumidity) );

	    if ( fHeatIndex <= tempF ){
		fHeatIndex = 0;
	    }
	}

	String unit = " C";
	pgrWetter.actInfo_TemperatureString = removeTrailingDotZero( decFormat.format(info_tempCFloat) ) + unit;
	pgrWetter.actInfo_DewPointString = removeTrailingDotZero( decFormat.format(info_dewCFloat) ) + unit;
	if (fHeatIndex >= 80){
	    pgrWetter.actInfo_HeatIndexString = removeTrailingDotZero( decFormat.format( TEMP_F_TO_C(fHeatIndex) ) ) + unit;
	}
    }

    @SuppressWarnings("static-access")
    private void calcWindChill(){
	//System.out.println("+++ calcWindChill() +++");
	float mathPowChill = (float) Math.pow(info_windMphFloat, 0.16d);
	float windChill = 35.74f +
		( 0.6215f * TEMP_C_TO_F(info_tempCFloat) ) -
		( 35.75f * mathPowChill ) +
		( 0.4275f * TEMP_C_TO_F(info_tempCFloat) * mathPowChill );

	if(windChill < 48){
	    pgrWetter.actInfo_WindChillString = decFormat.format(TEMP_F_TO_C(windChill));
	    pgrWetter.actInfo_WindChillString = removeTrailingDotZero(pgrWetter.actInfo_WindChillString) + " C";
	}
    }

    private String removeTrailingDotZero(String string) {
	if ( string.endsWith(",0") ) {
	    string = string.substring(0, string.length() - 2);
	}
	return string;
    }

    private float TEMP_C_TO_F(float x){
	return (x * 9 / 5) + 32;
    }

    private float TEMP_F_TO_C(float x){
	return (x - 32) * 5 / 9;
    }

    @SuppressWarnings("static-access")
    private void calcCurrentIcon(){
	//System.out.println("+++ calcCurrentIcon() +++");
	//System.out.println("+++ pgrWetter.actInfo_TheWeatherString: " + pgrWetter.actInfo_TheWeatherString + " +++");
	if( info_cloudsInt == -1 ){
	    info_cloudsInt = 64;
	}
	if(pgrWetter.actInfo_TheWeatherString.isEmpty()) {
	    if(info_cloudsInt == 0){
		pgrWetter.actInfo_TheWeatherString = iconName("sunny");
	    } else if(info_cloudsInt > 0 && info_cloudsInt <= 2){
		pgrWetter.actInfo_TheWeatherString = iconName("cloudy1");
	    } else if( info_cloudsInt > 2 && info_cloudsInt <= 4){
		pgrWetter.actInfo_TheWeatherString = iconName("cloudy2");
	    } else if( info_cloudsInt > 4 && info_cloudsInt <= 8){
		pgrWetter.actInfo_TheWeatherString = iconName("cloudy3");
	    } else if( info_cloudsInt > 8 && info_cloudsInt < 63){
		pgrWetter.actInfo_TheWeatherString = iconName("cloudy4");
	    } else {
		pgrWetter.actInfo_TheWeatherString = "cloudy5";
	    }
	} else if(pgrWetter.actInfo_TheWeatherString.equals("tstorm")) {
	    if( info_heavyBool ){
		info_cloudsInt = 30;
	    }
	    if(info_cloudsInt >= 0 && info_cloudsInt <= 10){
		pgrWetter.actInfo_TheWeatherString = iconName("tstorm1");
	    } else if( info_cloudsInt > 10 && info_cloudsInt <= 20){
		pgrWetter.actInfo_TheWeatherString = iconName("tstorm2");
	    } else {
		pgrWetter.actInfo_TheWeatherString = "tstorm3";
	    }
	} else if(pgrWetter.actInfo_TheWeatherString.equals("shower")) {
	    if( info_heavyBool ){
		info_cloudsInt = 30;
	    }
	    if(info_cloudsInt >= 0 && info_cloudsInt <= 10){
		pgrWetter.actInfo_TheWeatherString = iconName("shower1");
	    } else if( info_cloudsInt > 10 && info_cloudsInt <= 20){
		pgrWetter.actInfo_TheWeatherString = iconName("shower2");
	    } else {
		pgrWetter.actInfo_TheWeatherString = "shower3";
	    }
	} else if(pgrWetter.actInfo_TheWeatherString.equals("snow")) {
	    if( info_heavyBool ){
		info_cloudsInt = 30;
	    }
	    if(info_cloudsInt >= 0 && info_cloudsInt <= 8){
		pgrWetter.actInfo_TheWeatherString = iconName("snow1");
	    } else if( info_cloudsInt > 8 && info_cloudsInt <= 16){
		pgrWetter.actInfo_TheWeatherString = iconName("snow2");
	    } else if(info_cloudsInt > 16 && info_cloudsInt <= 24){
		pgrWetter.actInfo_TheWeatherString = iconName("snow3");
	    } else{
		pgrWetter.actInfo_TheWeatherString = "snow5";
	    }
	} else if(isNight(actStationID) && pgrWetter.actInfo_TheWeatherString.equals("mist")){
	    pgrWetter.actInfo_TheWeatherString = "mist_night";
	} else if(isNight(actStationID) && pgrWetter.actInfo_TheWeatherString.equals("fog")){
	    pgrWetter.actInfo_TheWeatherString = "fog_night";
	} else if( pgrWetter.actInfo_TheWeatherString.equals("mist") || pgrWetter.actInfo_TheWeatherString.equals("fog")) {
	    if ( info_cloudsInt >= 63 ){
		pgrWetter.actInfo_TheWeatherString = "cloudy5";
	    }
	}
//	    else {
//	    pgrWetter.actInfo_TheWeatherString = "dunno";
//	}
	//System.out.println("+++ pgrWetter.actInfo_TheWeatherString: " + pgrWetter.actInfo_TheWeatherString + " +++");
    }

    public final String getStationNameString(String stationID){
	String decodedString = getDecodedString(stationID);
	return decodedString.substring(0, decodedString.indexOf(stationID) + 5);
    }

    private final synchronized String getStringToParseString(String stationID){
	String StringToParseString = "";
        @SuppressWarnings("static-access")
	String infoURLString = pgrWetter.metarStationsString + stationID + ".TXT";
	try {
	    BufferedReader bufReader = new BufferedReader(
		    new InputStreamReader(
		    new URL(infoURLString).openStream()
		    )
		    );
	    String tmp = "";
	    while ( (tmp = bufReader.readLine() ) != null) {
		StringToParseString +=  " " + tmp;
	    }
	    bufReader.close();
	    pgrWetter.offline = false;
	} catch (MalformedURLException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getStringToParseString() MalformedURLException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	} catch (IOException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getStringToParseString() IOException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	}
	return StringToParseString.trim();
    }

    private final synchronized String getDecodedString(String stationID){
	String decodedString = "";
        @SuppressWarnings("static-access")
	String infoURLString = pgrWetter.metarDecodedString + stationID + ".TXT";
	try {
	    BufferedReader bufReader = new BufferedReader(
		    new InputStreamReader(
		    new URL(infoURLString).openStream()
		    )
		    );
	    String tmp = "";
	    while ( (tmp = bufReader.readLine() ) != null) {
		decodedString +=  "\n" + tmp;
	    }
	    bufReader.close();
	    pgrWetter.offline = false;
	} catch (MalformedURLException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getDecodedString() MalformedURLException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	} catch (IOException ex) {
	    //ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "getDecodedString() IOException", JOptionPane.ERROR_MESSAGE);
	    pgrWetter.offline = true;
	}
	return decodedString.trim();
    }

    public synchronized String[] updateLongLatVals(String stationID){
	//System.out.println("+++ updateLongLatVals(" + stationID + ") +++");

	String[] returnStringArray = {"", "", ""};
	// die Daten der Station holen
	decodedStringToParse = getDecodedString(stationID);

	// die Datenfelder in ein Stringarray packen und parsen
	//String[] parseData = decodedStringToParse.trim().split("\\s+"); // ### warum schneidet das das Minuszeichen ab?
	String[] parseDecodedData = decodedStringToParse.trim().split("[ ]+");

	for(int i = 0; i < parseDecodedData.length; i++){
	    String it = parseDecodedData[i];
	    //System.out.println("");
	    //System.out.println("### parseDecodedData[" + i + "]: '" + parseDecodedData[i] + "' ###");
	    if( Pattern.matches(findLat.pattern(), it) ){ // Latitude
		//System.out.println("### matches: '" + it + "' ###");
		returnStringArray[0] = it;
	    }
	    if( Pattern.matches(findLong.pattern(), it) ){ // Longitude
		//System.out.println("### matches: '" + it + "' ###");
		returnStringArray[1] = it;
	    }
//	    if( Pattern.matches(findHeight.pattern(), it) ){ // Hhenangabe
//		System.out.println("");
//		System.out.println("### matches: '" + it + "' ###");
//
//	    }
	}
	returnStringArray[2] = getStationNameString(stationID);
	return returnStringArray;
    }

    @SuppressWarnings("static-access")
    private boolean isNight(String stationID){
	//System.out.println("+++ isNight(" + stationID + ") +++");

	if ( actStationLat.equals("") || actStationLong.equals("") ){
	    return false;
	} else {
            @SuppressWarnings("static-access")
	    Sun theSun = new Sun( actStationLat, actStationLong, pgrWetter.actInfo_DateCal.getTime().getTime(), zenithModeInt );

            @SuppressWarnings("static-access")
	    long currently = (long) pgrWetter.actInfo_DateCal.getTime().getTime();
	    long civilStart = (long) theSun.computeSunRise();
	    long civilEnd = (long) theSun.computeSunSet();

	    pgrWetter.actInfo_sunRiseString = sdfUTC.format(new Date(civilStart));
	    pgrWetter.actInfo_sunSetString = sdfUTC.format(new Date(civilEnd));

	    //System.out.println("+++ currently:  " + currently + " +++");
	    //System.out.println("+++ civilStart: " + civilStart + " +++");
	    //System.out.println("+++ civilEnd:   " + civilEnd + " +++");

	    if( currently <= civilStart || currently >= civilEnd ){
		return true;
	    } else {
		return false;
	    }
	}
    }

    @SuppressWarnings("static-access")
    private boolean parseDate(String parseString){
	//System.out.println("+++ parseDate() +++");
	Matcher m = DateRegExp.matcher(parseString);
	if (m.matches()) {
	    String year = parseString.substring(0,4);
	    String month = parseString.substring(5,7);
	    String day = parseString.substring(8,10);
	    pgrWetter.actInfo_DateCal.set(Integer.parseInt(year), Integer.parseInt(month) -1, Integer.parseInt(day));
	    return true;
	} else {
	    return false;
	}
    }
    @SuppressWarnings("static-access")
    private boolean parseTime(String parseString){
	//System.out.println("+++ parseTime() +++");
	Matcher m = TimeRegExp.matcher(parseString);
	if (m.matches()) {
            @SuppressWarnings("static-access")
	    int year = pgrWetter.actInfo_DateCal.get(Calendar.YEAR);
            @SuppressWarnings("static-access")
	    int month = pgrWetter.actInfo_DateCal.get(Calendar.MONTH);
            @SuppressWarnings("static-access")
	    int day = pgrWetter.actInfo_DateCal.get(Calendar.DAY_OF_MONTH);
	    String hour = parseString.substring(0,2);
	    String minute = parseString.substring(3,5);
	    pgrWetter.actInfo_DateCal.set(year, month, day, Integer.parseInt(hour), Integer.parseInt(minute));
	    return true;
	} else {
	    return false;
	}
    }
    @SuppressWarnings("static-access")
    private boolean parseWindSpeed(String parseString){
	Matcher m = WindRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("--- parseWindSpeed(" + parseString + ") ---");

	    int direction;
	    float windSpeed;
	    float gustSpeed;
	    String sWindUnit;

	    if(m.group(1) != null){ direction = Integer.parseInt( m.group(1) ); } else { direction = 0; }
	    if(m.group(2) != null){ windSpeed = Float.parseFloat( m.group(2) ); } else { windSpeed = 0f; }
	    if(m.group(3) != null){ gustSpeed = Float.parseFloat( m.group(3) ); } else { gustSpeed = 0f; }
	    if(m.group(4) != null){ sWindUnit = m.group(4); } else { sWindUnit = ""; }

	    if (sWindUnit.equals("KT") ) {
		windSpeed = (windSpeed * 1.85f);
		gustSpeed = (gustSpeed * 1.85f);
	    } else if (sWindUnit.equals("MPS") ) {
		windSpeed = (windSpeed * 3.6f);
		gustSpeed = (gustSpeed * 3.6f);
	    }
	    info_windMphFloat = (windSpeed / 1.6f);
	    pgrWetter.actInfo_WindSpeedString = decFormat.format(windSpeed) + " km/h (" + decVKFormat.format(windSpeed / 1.85) + " kt)";

	    if (gustSpeed >= 1) {
		pgrWetter.actInfo_CurrentVect.addElement("Ben bis zu " + (int) gustSpeed + " km/h (" + decVKFormat.format(gustSpeed / 1.85) + " kt)");
	    }

	    if ( !m.group(1).equals("VRB") && (windSpeed >= 1)) {
		pgrWetter.actInfo_WindDirectionString = parseWindDirection(direction);
	    }
	    return true;
	} else {
	    return false;
	}
    }

    private String parseWindDirection(int direction){
	//System.out.println("--- parseWindDirection(" + direction + ") ---");
	String returnString = "";
	for (int i = 0; i < wind_directions.length - 1; i++) {
	    if (direction >= Integer.parseInt(wind_directions[i][0]) && direction < Integer.parseInt(wind_directions[i + 1][0]) ){
		returnString = wind_directions[i][1];
		break;
	    }
	}
	return returnString;
    }

    @SuppressWarnings("static-access")
    private boolean parseVisibility(String parseString){
	float fVisibility = 0;

	Matcher m = VisRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("+++ parseVisibility() +++");
	    fVisibility = Float.parseFloat(parseString);
	}
	if (fVisibility > 0){
	    if(fVisibility == 9999){
		pgrWetter.actInfo_VisibilityString = "ber 10 km";
	    } else {
		fVisibility *= 0.001;
		pgrWetter.actInfo_VisibilityString = String.valueOf(fVisibility) + " km";
	    }
	    return true;
	} else	{
	    return false;
	}
    }

    private boolean parseTemperature(String parseString){
	Matcher m = TempRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("+++ parseTemperature() +++");

	    float fTemp = 0f;
	    float fDew = 0f;

	    if(m.group(2) != null){ fTemp = Float.parseFloat( m.group(2) ); }
	    if(m.group(4) != null){ fDew = Float.parseFloat( m.group(4) ); }

	    if ( ( m.group(1) != null && m.group(1).equals("M") ) && fTemp != 0 ){
		fTemp *= -1;
	    }
	    if ( ( m.group(3) != null && m.group(3).equals("M") ) && fDew != 0 ){
		fDew *= -1;
	    }
	    info_tempCFloat = fTemp;
	    info_dewCFloat = fDew;
	    return true;
	} else {
	    return false;
	}
    }
    @SuppressWarnings("static-access")
    private boolean parsePressure(String parseString){
	Matcher m = PressRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("--- parsePressure(" + parseString + ") ---");

	    String type = "";
	    float fPressure = 0f;

	    if(m.group(1) != null){ type = m.group(1); }
	    if(m.group(2) != null){ fPressure = Float.parseFloat( m.group(2) ); }

	    if (type.equals("A") ){
		pgrWetter.actInfo_PressureString = decVKFormat.format(fPressure * (33.8639 / 100) ) + " hPa (" + decFormat.format(fPressure) + " Hg)";
	    } else if (type.equals("Q") ){
		pgrWetter.actInfo_PressureString = decVKFormat.format(fPressure) + " hPa (" + decFormat.format(fPressure / 33.8639) + " Hg)";
	    }
	    return true;
	} else {
	    return false;
	}
    }
    @SuppressWarnings("static-access")
    private boolean parseCover(String parseString){
	Matcher m = CoverRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("+++ parseCover(" + parseString + ") +++");

	    String sCode = "";
	    float height = 0f;
	    String sClouds = "";
	    String skycondition = "";

	    if(m.group(1) != null){ sCode = m.group(1); } else { sCode = ""; }
	    if(m.group(2) != null){ height = Float.parseFloat(m.group(2)); } else { height = 0f; }
	    //System.out.println("+++ m.group(1): " + m.group(1) + " +++");
	    //System.out.println("+++ m.group(2): " + m.group(2) + " +++");

	    height *= 100;
	    height = height * 0.3048f;
	    sClouds = (int)height + " Metern";

	    if ( sCode.equals("FEW") ) {
		skycondition = "Schwache Bewlkung in einer Hhe von " + sClouds;
		info_cloudsInt += 2;
	    } else if (sCode.equals("SCT") ) {
		skycondition = "Vereinzelte Wolken in einer Hhe von " + sClouds;
		info_cloudsInt += 4;
	    } else if (sCode.equals("BKN") ) {
		skycondition = "Aufgelockerte Bewlkung in einer Hhe von " + sClouds;
		info_cloudsInt += 8;
	    } else if (sCode.equals("OVC") ) {
		skycondition = "Geschlossene Bewlkung in einer Hhe von " + sClouds;
		info_cloudsInt += 64;
	    } else if ( sCode.equals("CLR") || sCode.equals("SKC") || sCode.equals("CAVOK") ) {
		skycondition = "Klarer Himmel";
		pgrWetter.actInfo_VisibilityString = "ber 10 km";
		info_cloudsInt = 0;
	    }
	    pgrWetter.actInfo_CoverVect.addElement(skycondition);
	    return true;
	} else {
	    return false;
	}
    }
    @SuppressWarnings("static-access")
    private boolean parseCurrent(String parseString){
	Matcher m = CurrentRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("--- parseCurrent(" + parseString + ") ---");

	    String sIntensity = "";
	    String sCode = "";
	    String intensity = "";
	    String descriptor = "";
	    String phenomena = "";
	    String currentWeather = "";

	    if(m.group(1) != null){ sIntensity = m.group(1); }
	    if(m.group(2) != null){ sCode = m.group(2); }
	    //System.out.println("--- sIntensity -> m.group(1): " + m.group(1) + " ---");
	    //System.out.println("--- sCode -> m.group(2): " + m.group(2) + " ---");

	    // Decode the intensity
	    if (sIntensity.equals("+") ){
		intensity = "starker";
		info_heavyBool = true;
	    } else if (sIntensity.equals("-") ){
		intensity = "leichter";
		info_heavyBool = false;
	    } else if (sIntensity.equals("") ){
		intensity = "miger";
		info_heavyBool = false;
	    }

	    // Decode the descriptor
	    if (sCode.contains("MI")){
		descriptor = "Oberflchlich";
	    } else if (sCode.contains("PR")){
		descriptor = "Teilweise";
	    } else if (sCode.contains("BC")){
		descriptor = "Fleckenweise";
	    } else if (sCode.contains("DR")){
		descriptor = "Niedrig ziehend";
	    } else if (sCode.contains("BL")){
		descriptor = "Strmisch";
	    } else if (sCode.contains("SH")){
		descriptor = "Schauer";
		pgrWetter.actInfo_TheWeatherString = "shower";
	    } else if (sCode.contains("TS")){
		descriptor = "Gewitter";
		pgrWetter.actInfo_TheWeatherString = "tstorm";
	    } else if (sCode.contains("FZ")){
		descriptor = "Frost";
	    }

	    // Decode weather phenomena
	    if (sCode.contains("DZ")){
		phenomena = "Nieselregen";
		pgrWetter.actInfo_TheWeatherString = "light_rain";
	    } else if (sCode.contains("RA")){
		phenomena = "Regen";
		pgrWetter.actInfo_TheWeatherString = "shower";
	    } else if (sCode.contains("SN")){
		phenomena = "Schnee";
		pgrWetter.actInfo_TheWeatherString = "snow";
	    } else if (sCode.contains("SG")){
		phenomena = "Graupel";
		pgrWetter.actInfo_TheWeatherString = "snow4";
	    } else if (sCode.contains("IC")){
		phenomena = "Eiskristalle";
		pgrWetter.actInfo_TheWeatherString = "hail";
	    } else if (sCode.contains("PE")){
		phenomena = "Eiskrner";
		pgrWetter.actInfo_TheWeatherString = "hail";
	    } else if (parseString.contains("GR")){
		phenomena = "Hagel";
		pgrWetter.actInfo_TheWeatherString = "hail";
	    } else if (sCode.contains("GS")){
		phenomena = "Kleine Hagelkrner";
		pgrWetter.actInfo_TheWeatherString = "hail";
	    } else if (parseString.contains("UP")){
		phenomena = "Unbekannter Niederschlag";
		pgrWetter.actInfo_TheWeatherString = "shower1";
	    } else if (sCode.contains("BR")){
		phenomena = "Dunst";
		// Mist has lower priority than say rain or snow
		if ( pgrWetter.actInfo_TheWeatherString.isEmpty() ){
		    pgrWetter.actInfo_TheWeatherString = "mist";
		}
	    } else if (sCode.contains("FG")){
		phenomena = "Nebel";
		// Fog has lower priority than say rain or snow
		if ( pgrWetter.actInfo_TheWeatherString.isEmpty() ){
		    pgrWetter.actInfo_TheWeatherString = "fog";
		}
	    } else if (sCode.contains("FU")){
		phenomena = "Rauch";
	    } else if (sCode.contains("VA")){
		phenomena = "Vulkanasche";
	    } else if (sCode.contains("DU")){
		phenomena = "Weitverbreiteter Staub";
	    } else if (sCode.contains("SA")){
		phenomena = "Sand";
	    } else if (sCode.contains("HZ")){
		phenomena = "Dunst";
	    } else if (sCode.contains("PY")){
		phenomena = "Sprhregen";
	    } else if (sCode.contains("PO")){
		phenomena = "Staub/Sandwirbel";
	    } else if (sCode.contains("SQ")){
		phenomena = "Pltzliche Ben";
	    } else if (sCode.contains("FC")){
		if (sIntensity.equals("+")){
		    currentWeather = "Tornado";
		} else {
		    phenomena = "Luftwirbel";
		}
	    } else if (sCode.contains("SS")){
		phenomena = "Sandsturm";
	    } else if (sCode.contains("DS")){
		phenomena = "Staubsturm";
	    }

	    if (currentWeather.length() < 1){
		currentWeather = intensity + " " + phenomena + " " + descriptor;
//		currentWeather = intensity + " ist die Intensitt, " +
//			descriptor + " ist die Beschreibung und " +
//			phenomena + " ist das Phnomen";
	    }

	    if ( currentWeather.length() > 0 ){
		pgrWetter.actInfo_CurrentVect.addElement(currentWeather);
	    }
	    //System.out.println("--- currentWeather: " + currentWeather + " ---");
	    return true;
	} else {
	    return false;
	}
    }
    private boolean parseTemperatureTenths(String parseString){
	Matcher m = TempTenRegExp.matcher(parseString);
	if (m.matches()) {
	    //System.out.println("+++ parseTemperatureTenths() +++");

	    float temperature = 0f;
	    float dewPoint = 0f;

	    if(m.group(1) != null){ temperature = Float.parseFloat( m.group(1) ) / 10; }
	    if(m.group(2) != null){ dewPoint = Float.parseFloat( m.group(2) ) / 10; }

	    if ( temperature >= 100 ){
		temperature -= 100;
		temperature *= -1;
	    }
	    if ( dewPoint >= 100 ){
		dewPoint -= 100;
		dewPoint *= -1;
	    }
	    info_tempCFloat = temperature;
	    info_dewCFloat = dewPoint;

	    //System.out.println("+++ info_tempCFloat: " + info_tempCFloat + " +++");
	    //System.out.println("+++ info_dewCFloat: " + info_dewCFloat + " +++");
	    return true;
	} else {
	    return false;
	}
    }
    private boolean parseStationNeedsMaintenance(String parseString){
	Matcher m = StatNeedsMaint.matcher(parseString);
	if (m.matches()) {
	    info_stationNeedsMaintBool = true;

	    //System.out.println("+++ info_stationNeedsMaintBool: " + info_stationNeedsMaintBool + " +++");
	    return true;
	} else {
	    return false;
	}
    }

    private String iconName(String iconName){
	//System.out.println("+++ isNight(actStationID): " + isNight(actStationID) + " +++");
	if ( isNight(actStationID) ){
	    iconName += "_night";
	}
	return iconName;
    }

    private synchronized void createStationsFull() {
	String text = new String();
	String line;
	try {
	    InputStream is = Main.class.getResource("/PGRWetter/stations/stations.dat").openStream();
	    BufferedReader reader = new BufferedReader( new InputStreamReader(is) );
	    String[] felder;
	    while((line = reader.readLine()) != null) {
		felder = line.split(";");
		stationsFull.put(felder[0], felder[3] + " | " + felder[5]);
	    }
	    reader.close();
	    is.close();
	} catch (IOException ex) {
	    ex.printStackTrace();
	    //JOptionPane.showMessageDialog(null, ex.toString(), "createStationsFull() IOException", JOptionPane.ERROR_MESSAGE);
	}

    }
}
